#lang racket/base

(require racket/contract)

(provide
  (contract-out
    [group-tasks (->* ((listof task?))
                      ((listof task-group-spec?) (listof task-sort-spec?))
                      (or/c (listof task-group?) (listof task?)))]
    [group-tasks-result->string (->* ((or/c (listof task-group?) (listof task?)))
                                     (#:omit-fields valid-field-omissions?)
                                     string?)]
    (struct task-group ([title (or/c string? #f)]
                        [contents (or/c (listof task-group?) (listof task?))]))
    (struct task-group-spec ([order sort-order/c]
                             [field task-field/c]
                             [tag-key (or/c string? #f)])))
)

(require
  racket/list
  racket/string
  "task.rkt")


; --------------------------------------------------------------------------------
; Groups of tasks. See also `task.rkt`.

; Order-symbol Task-field-symbol [String | #f] -> Task-group-spec
(struct task-group-spec task-sort-spec
  ; `tag-key` is a string for the tag key, say, "due". `#f` if `field` isn't a
  ; 'tags field.
  (tag-key)
  #:transparent)

; [String | #f] [[List-of Task-group] | [List-of Task]] -> Task-group
(struct task-group
  (; Group title, for example "+todoreport". `#f` if no title should be used.
   title
   ; List of `task`s or `task-group`s
   contents)
  #:transparent)


; --------------------------------------------------------------------------------
; Define helper functions for grouping tasks.

; There are the following types of fields:
; - the `completed?` field
; - simple fields (either a string or `#f`)
; - list fields (projects or contexts)
; - tag field
;
; For each of these field types define a function to return the titles for all
; groups that should be created, given a task group spec and a task.

; --------------------------------------------------
; `completed?` field

; Task-group-spec Task -> [List-of Title-string]
;
; Return a list with a title string for the `task`, given a task group spec.
;
; Example:
;   (titles-for-task/completed? (task-group-spec 'asc 'completed? #f) some-task)
;   -> '("Not completed")
(define (titles-for-task/completed? group-spec task)
  (define completed? (eq? (task-field task 'completed?) #t))
  (if completed?
      '("Completed")
      (list (group-params-value 'completed? 'unset-value-title))))

; --------------------------------------------------
; Simple string fields

; Task-group-spec Task -> [List-of Title-string]
;
; Return a list with a title string for the `task`, given a task group spec.
;
; Example:
;   (titles-for-task/string-field (task-group-spec 'asc 'creation-date #f) some-task)
;   -> '("2021-03-16")
(define (titles-for-task/string-field group-spec task)
  (define field-value (task-field task group-spec))
  (list
    (if (eq? field-value #f)
        (group-params-value (task-sort-spec-field group-spec) 'unset-value-title)
        field-value)))

; --------------------------------------------------
; List fields

; Task-group-spec Task -> [List-of Title-string]
;
; Return a list with possible title strings for the `task`, given a task group
; spec.
;
; Example:
;   (titles-for-task/list-field (task-group-spec 'asc 'contexts #f) some-task)
;   -> '("@home" "@work")
(define (titles-for-task/list-field group-spec task)
  (define field-symbol (task-sort-spec-field group-spec))
  (define field-value (task-field task field-symbol))
  (if (empty? field-value)
      (list (group-params-value field-symbol 'unset-value-title))
      (map (group-params-value field-symbol 'field-value->title) field-value)))

; --------------------------------------------------
; Tags fields

; Task-group-spec Task -> [List-of Title-string]
;
; Return a list with title strings for the `task`, given a task group spec.
;
; Example:
;   (titles-for-task/tags-field (group-spec 'asc 'tags "due") some-task)
;   -> '("due:2021-03-18")
;
; Note that a task can have multiple values for the same key, for example if
; the task string is "foo:bar foo:baz".
(define (titles-for-task/tags-field group-spec task)
  (define field-symbol (task-sort-spec-field group-spec))
  (define field-value (task-field task field-symbol))
  (define tag-key (task-group-spec-tag-key group-spec))
  (define pairs (task-tags task))
  (define pair-titles
    (for/list ([pair pairs]
               #:when (string=? (car pair) tag-key))
      ((group-params-value field-symbol 'field-value->title) pair)))
  (if (null? pair-titles)
      (list (group-params-value field-symbol 'unset-value-title))
      pair-titles))


; --------------------------------------------------------------------------------
; Define parameters for grouping by different task fields.

; [Hash-of Task-symbol Group-params]
(define group-params
  (hasheq
    'completed?
      (hasheq 'unset-value-title "Not completed"
              'field-value->title (lambda (value) "Completed")
              'titles-for-task titles-for-task/completed?)
    'priority
      (hasheq 'unset-value-title "No priority"
              'field-value->title values
              'titles-for-task titles-for-task/string-field)
    'completion-date
      (hasheq 'unset-value-title "No completion date"
              'field-value->title values
              'titles-for-task titles-for-task/string-field)
    'creation-date
      (hasheq 'unset-value-title "No creation date"
              'field-value->title values
              'titles-for-task titles-for-task/string-field)
    'description
      (hasheq 'unset-value-title "No description"
              'field-value->title values
              'titles-for-task titles-for-task/string-field)
    'projects
      (hasheq 'unset-value-title "No project"
              'field-value->title ; This will be called on each of the list items in
                                   ; the `projects` field.
                                   (lambda (value) (string-append "+" value))
              'titles-for-task titles-for-task/list-field)
    'contexts
      (hasheq 'unset-value-title "No context"
              'field-value->title ; This will be called on each of the list items in
                                   ; the `contexts` field.
                                   (lambda (value) (string-append "@" value))
              'titles-for-task titles-for-task/list-field)
    ; Tags need special handling. We want to group by tag/value combinations,
    ; not just tag names.
    'tags
      (hasheq 'unset-value-title "No such tag"
              'field-value->title ; This will be called on each hash key/value pair
                                   ; for the selected key.
                                   (lambda (key-value-pair)
                                     (string-append
                                       (car key-value-pair)
                                       ":"
                                       (cdr key-value-pair)))
              'titles-for-task titles-for-task/tags-field)))

; [X]  Task-field-symbol Property-symbol -> X
;
; Return a `group-params` value for the property for a task field.
;
; Example:
;   (group-params-value 'priority 'unset-value-title)
;   -> "No priority"
(define (group-params-value task-field property)
  (hash-ref (hash-ref group-params task-field) property))

; Task Task-group-spec -> [List-of String]
;
; Return list of possible titles for the task according to `group-spec`.
; The titles aren't in any particular order.
;
; Example:
;   (titles-for-task (string->task "Test task @home @computer +todo.txt")
;                    (task-group-spec 'asc 'context #f))
;   -> '("@home" "@computer")
(define (titles-for-task task group-spec)
  (define titles-for-task-func
    (group-params-value (task-sort-spec-field group-spec) 'titles-for-task))
  (titles-for-task-func group-spec task))


; --------------------------------------------------------------------------------
; Group tasks.

; [List-of Task] Task-group-spec -> [List-of Title-string]
;
; Return all possible title strings for a list of tasks, given a task group spec.
(define (titles-for-tasks tasks group-spec)
  (remove-duplicates
    (append*
      (for/list ([task tasks])
        (titles-for-task task group-spec)))))

; Title-string [List-of Task] Task-group-spec -> [List-of Task]
;
; Return a list of tasks which belong to the given title, given a list of
; tasks.
;
; Example:
;   (tasks-for-title "@home"
;                    (list task1 task2 task3)
;                    (task-group-spec 'asc 'contexts #f))
;   -> (list task1 task3)
(define (tasks-for-title title tasks group-spec)
  (define (task-belongs-to-title? task)
    (define task-titles (titles-for-task task group-spec))
    (index-of task-titles title))
  (filter task-belongs-to-title? tasks))

; Task-group-spec -> [String String -> Boolean]
;
; Return a function to be used for sorting task groups titles, i.e. in the
; `sort` function in the standard library.
(define (make-task-titles-sort-func group-spec)
  (define task-field (task-sort-spec-field group-spec))
  (define unset-value-title (group-params-value
                              (task-sort-spec-field group-spec)
                              'unset-value-title))
  (define sort-order (task-sort-spec-order group-spec))
  (lambda (group1-title group2-title)
    (cond
      ; For the `completed?` task field and ascending sort order, we want to
      ; sort the non-completed tasks _before_ the completed tasks. That is, we
      ; want the opposite of the default behavior of sorting the tasks with the
      ; "unset" title last. The sort order for this special case is the same as
      ; for `sort-tasks`. The background is that uncompleted tasks have their
      ; `'completed?` field set to `#f`, and we want to sort `#f` before `#t`.
      [(and (eq? task-field 'completed?) (eq? sort-order 'asc))
       (cond
         [(string=? group1-title unset-value-title) #t]
         [(string=? group2-title unset-value-title) #f])]
      ; Sort `unset-value-title` to the end. There's no need to consider a
      ; combination of _two_ `unset-value-title` since only one group can have
      ; this title.
      ;
      ; As long as the `unset-value-title` isn't a capital letter (for
      ; priority), a date (completion and creation date), doesn't start with
      ; `@` or `+` (contexts and projects) or contains a colon (tag and its
      ; values), this case should never happen. That said, it would probably be
      ; more robust if we could use a special non-string value for the unset
      ; title and only replace it later (after sorting) with the actual title.
      [(string=? group1-title unset-value-title) #f]
      [(string=? group2-title unset-value-title) #t]
      ; Otherwise do a normal string comparison.
      [(eq? sort-order 'asc)  (string<? group1-title group2-title)]
      [(eq? sort-order 'desc) (string>? group1-title group2-title)]
      [else (error "should never be reached")])))

; [List-of Task] [List-of Task-group-spec] [List-of Task-sort-spec]
; -> [[List-of Task-group] | [List-of Task]]
;
; Return the `tasks` grouped according to the `group-specs` and sorted
; according to `sort-specs`.
;
; If `group-specs` is not empty, the result is a list of task groups, otherwise
; the result is list of tasks.
;
; Groups can be nested. The outermost grouping level is taken from the first
; `task-group-spec` in `group-specs`. The innermost level of the result
; consists of one or more groups, where each group contains a lists of tasks,
; sorted according to `sort-specs`. See the documentation of `sort-tasks` (from
; `task.rkt`) for how the sort specs are applied.
;
; Example:
;
;   (group-tasks
;     (map string->task '("2021-12-11 Early task @home"
;                         "2021-12-11 Early task without contexts"
;                         "2021-12-12 Late task @work"
;                         "2021-12-11 Early task @work"
;                         "2021-12-12 Late task @home @phone"))
;     (list (task-group-spec 'asc 'context #f)
;           (task-group-spec 'desc 'creation-date #f)))
;
; gives
;
;   (list
;     (task-group "@home"
;       (list
;         (task-group "2021-12-12"
;           (list (string->task "Late task @home @phone")))
;         (task-group "2021-12-11"
;           (list (string->task "Early task @home")))))
;     (task-group "@phone"
;       (list
;         (task-group "2021-12-12"
;           (list (string->task "Late task @home @phone")))))
;     (task-group "@work"
;       (list
;         (task-group "2021-12-12"
;           (list (string->task "Late task @work")))
;         (task-group "2021-12-11"
;           (list (string->task "Early task @work")))))
;     (task-group "No context"
;       (list
;         (task-group "2021-12-11"
;           (list (string->task "Early task without contexts"))))))
;
; or, in text form,
;
;   @home
;     2021-12-12
;       Late task @home @phone
;     2021-12-11
;       Early task @home
;   @work
;     2021-12-12
;       Late task @work
;     2021-12-11
;       Early task @work
;   No context
;     2021-12-11
;       Early task without contexts
(define (group-tasks tasks [group-specs '()] [sort-specs '()])
  (cond
    [(empty? group-specs)
     (sort-tasks tasks sort-specs)]
    [else
     (define sorted-titles
       (sort (titles-for-tasks tasks (car group-specs))
             (make-task-titles-sort-func (car group-specs))))
     (for/list ([group-title sorted-titles])
       (define group-items
         (group-tasks
           ; Tasks for this group.
           (tasks-for-title group-title tasks (car group-specs))
           (cdr group-specs)
           sort-specs))
       (task-group group-title group-items))]))


; --------------------------------------------------------------------------------
; Convert a `group-tasks` result to a string.

; String Integer -> String
;
; Return the string `str` where each line is indented by `indent-level`
; indents.
(define (indent-string str indent-level)
  (define indent-string-per-level "  ")
  (define indent (string-append* (make-list indent-level indent-string-per-level)))
  (string-append indent str))

; [[List-of Task-group] | [List-of Task]] #:omit-fields [List-of Task-field-symbol]-> String
;
; Return a multiline string representing the result of a previous `group-tasks`
; call.
;
; The string uses nesting to show the grouping. An example result might look
; like this:
;
;   Group 1 title
;     Group 1.1 title
;       Task 1.1.1
;       Task 1.1.2
;       Task 1.1.3
;     Group 1.2 title
;       Task 1.2.1
;       Task 1.2.2
;       Task 1.2.3
;   Group 2 title
;     Group 2.1 title
;       Task 2.1.1
;       Task 2.1.2
;       Task 2.1.3
;     Group 2.2 title
;       Task 2.2.1
;       Task 2.2.2
;       Task 2.2.3
;
; `#:omit-fields` can be a list of fields to omit from each task string. Only
; simple fields, i.e. 'completed?, 'priority, 'completion-date and
; 'creation-date are allowed. Omitting other data would be more complicated and
; could return odd-looking task strings (as in "Finish article for +my_project"
; -> "Finish article for").
;
; It's an error to try to omit the creation date, but not the completion date.
; This is because this would leave the completion date as the single date, but
; a single date is supposed to be a creation date.
(define (group-tasks-result->string grouping-result #:omit-fields [omit-fields '()])
  (define lines
    (let loop ([grouping-result grouping-result]
               [indent-level 0])
      (cond
        [(null? grouping-result)
         '()]
        [(task? (car grouping-result))
         (define (task->indented-string task)
           (indent-string (task->string task #:omit-fields omit-fields) indent-level))
         (map task->indented-string grouping-result)]
        [(task-group? (car grouping-result))
         (define (group->indented-strings group)
           (cons (indent-string (task-group-title group) indent-level)
                 (loop (task-group-contents group) (add1 indent-level))))
         (append*
           (map group->indented-strings grouping-result))])))
  (string-join lines "\n"))


; --------------------------------------------------------------------------------

(module+ test
  (require
    rackunit
    al2-test-runner
    (submod ".."))

  ; Short for "task string list"
  (define (tsl . strings) (map string->task strings))
  (define tg task-group)

  (define-check (check-grouping
                 before-tasks
                 task-group-specs
                 sort-specs
                 expected-grouping-result)
    (define grouping-result (group-tasks before-tasks task-group-specs sort-specs))
    (check-equal? grouping-result expected-grouping-result))

  (run-tests
    (test-suite "Grouping"
      ; If no group specs and sort specs are given, the task list should be
      ; returned unchanged.
      (test-case "No group specs"
        (check-grouping
          (tsl "2021-03-02 Task 1"
               "2021-03-03 Task 2")
          '()
          '()
          (tsl "2021-03-02 Task 1"
               "2021-03-03 Task 2")))
      (test-case "Group by ascending creation date"
        (check-grouping
          (tsl "2021-02-01 Task 1"
               "2021-02-03 Task 2"
               "Task 3"
               "2021-02-01 Task 4")
          (list (task-group-spec 'asc 'creation-date #f))
          '()
          (list (tg "2021-02-01"       (tsl "2021-02-01 Task 1"
                                            "2021-02-01 Task 4"))
                (tg "2021-02-03"       (tsl "2021-02-03 Task 2"))
                (tg "No creation date" (tsl "Task 3")))))
      (test-case "Group by descending creation date"
        (check-grouping
          (tsl "2021-02-01 Task 1"
               "2021-02-03 Task 2"
               "Task 3"
               "2021-02-01 Task 4")
          (list (task-group-spec 'desc 'creation-date #f))
          '()
          (list (tg "2021-02-03"       (tsl "2021-02-03 Task 2"))
                (tg "2021-02-01"       (tsl "2021-02-01 Task 1"
                                            "2021-02-01 Task 4"))
                (tg "No creation date" (tsl "Task 3")))))
      (test-case "Group by ascending completion"
        (check-grouping
          (tsl "x Completed 2"
               "Not completed 1"
               "Not completed 2"
               "x Completed 1")
          (list (task-group-spec 'asc 'completed? #f))
          '()
          (list (tg "Not completed" (tsl "Not completed 1"
                                         "Not completed 2"))
                (tg "Completed"     (tsl "x Completed 2"
                                         "x Completed 1")))))
      (test-case "Group by descending completion"
        (check-grouping
          (tsl "x Completed 2"
               "Not completed 1"
               "Not completed 2"
               "x Completed 1")
          (list (task-group-spec 'desc 'completed? #f))
          '()
          (list (tg "Completed"     (tsl "x Completed 2"
                                         "x Completed 1"))
                (tg "Not completed" (tsl "Not completed 1"
                                         "Not completed 2")))))
      (test-case "Group by ascending project"
        (check-grouping
          (tsl "2021-02-01 Task with +project1"
               "2021-02-01 Task without projects"
               "2021-02-03 Task with +project1 +project2")
          (list (task-group-spec 'asc 'projects #f))
          '()
          (list (tg "+project1"  (tsl "2021-02-01 Task with +project1"
                                      "2021-02-03 Task with +project1 +project2"))
                (tg "+project2"  (tsl "2021-02-03 Task with +project1 +project2"))
                (tg "No project" (tsl "2021-02-01 Task without projects")))))
      (test-case "Group by descending project"
        (check-grouping
          (tsl "2021-02-01 Task with +project1"
               "2021-02-01 Task without projects"
               "2021-02-03 Task with +project1 +project2")
          (list (task-group-spec 'desc 'projects #f))
          '()
          (list (tg "+project2"  (tsl "2021-02-03 Task with +project1 +project2"))
                (tg "+project1"  (tsl "2021-02-01 Task with +project1"
                                      "2021-02-03 Task with +project1 +project2"))
                (tg "No project" (tsl "2021-02-01 Task without projects")))))
      (test-case "Group by ascending tag value"
        (check-grouping
          (tsl "2021-02-01 Task due:2021-03-20"
               "2021-02-01 Task with other tag other:foo"
               "2021-02-03 Task without tag"
               "2021-02-03 Task with other due:2021-03-18")
          (list (task-group-spec 'asc 'tags "due"))
          '()
          (list (tg "due:2021-03-18" (tsl "2021-02-03 Task with other due:2021-03-18"))
                (tg "due:2021-03-20" (tsl "2021-02-01 Task due:2021-03-20"))
                (tg "No such tag"    (tsl "2021-02-01 Task with other tag other:foo"
                                          "2021-02-03 Task without tag")))))
      (test-case "Group by descending tag value"
        (check-grouping
          (tsl "2021-02-01 Task due:2021-03-20"
               "2021-02-01 Task with other tag other:foo"
               "2021-02-03 Task without tag"
               "2021-02-03 Task with other due:2021-03-18")
          (list (task-group-spec 'desc 'tags "due"))
          '()
          (list (tg "due:2021-03-20" (tsl "2021-02-01 Task due:2021-03-20"))
            (tg "due:2021-03-18"     (tsl "2021-02-03 Task with other due:2021-03-18"))
            (tg "No such tag"        (tsl "2021-02-01 Task with other tag other:foo"
                                          "2021-02-03 Task without tag")))))
      (test-case "Group by ascending tag value for multiple tag values for the same key"
        (check-grouping
          (tsl "2021-02-01 Task foo:bar"
               "2021-02-01 Task with other tag other:foo"
               "2021-02-03 Task without tag"
               "2021-02-03 Task with other foo:baz foo:bar")
          (list (task-group-spec 'asc 'tags "foo"))
          '()
          (list (tg "foo:bar"     (tsl "2021-02-01 Task foo:bar"
                                       "2021-02-03 Task with other foo:baz foo:bar"))
                (tg "foo:baz"     (tsl "2021-02-03 Task with other foo:baz foo:bar"))
                (tg "No such tag" (tsl "2021-02-01 Task with other tag other:foo"
                                       "2021-02-03 Task without tag")))))
      (test-case "Group by descending tag value for multiple tag values for the same key"
        (check-grouping
          (tsl "2021-02-01 Task foo:bar"
               "2021-02-01 Task with other tag other:foo"
               "2021-02-03 Task without tag"
               "2021-02-03 Task with other foo:baz foo:bar")
          (list (task-group-spec 'desc 'tags "foo"))
          '()
          (list (tg "foo:baz"     (tsl "2021-02-03 Task with other foo:baz foo:bar"))
                (tg "foo:bar"     (tsl "2021-02-01 Task foo:bar"
                                       "2021-02-03 Task with other foo:baz foo:bar"))
                (tg "No such tag" (tsl "2021-02-01 Task with other tag other:foo"
                                       "2021-02-03 Task without tag")))))
      ; If both group specs and sort specs are given, the sort specs should be
      ; applied to each group.
      (test-case "Grouping and sorting"
        (check-grouping
          (tsl "2021-03-12 Task without context"
               "Task without dates and context"
               "2021-03-13 Task @home @work"
               "2021-03-11 Another task without context"
               "2021-03-15 Task @work"
               "Task without date @home")
          (list (task-group-spec 'asc 'contexts #f))
          (list (task-sort-spec 'desc 'creation-date))
          (list (tg "@home"      (tsl "2021-03-13 Task @home @work"
                                      "Task without date @home"))
                (tg "@work"      (tsl "2021-03-15 Task @work"
                                      "2021-03-13 Task @home @work"))
                (tg "No context" (tsl "2021-03-12 Task without context"
                                      "2021-03-11 Another task without context"
                                      "Task without dates and context")))))
    ; If no sort specs are given, a default of '() should be assumed.
    (test-case "Default sort specs, i.e. an empty list"
      (define before-tasks (tsl "2021-02-01 Task 1"
                                "2021-02-03 Task 2"
                                "2021-02-01 Task 3"))
      (define task-group-specs (list (task-group-spec 'asc 'creation-date #f)))
      (define grouping-result (group-tasks before-tasks task-group-specs))
      (define expected-grouping-result
        (list (tg "2021-02-01" (tsl "2021-02-01 Task 1"
                                    "2021-02-01 Task 3"))
              (tg "2021-02-03" (tsl "2021-02-03 Task 2"))))
      (check-equal? grouping-result expected-grouping-result))
    (test-case "Nested grouping with ascending completion"
      (check-grouping
        (tsl "2021-12-11 Incomplete task 1"
             "x 2021-12-12 Complete task 1"
             "Task without creation date"
             "2021-12-12 Incomplete task 2"
             "2021-12-11 Incomplete task 3"
             "x 2021-12-11 Complete task 2")
        (list (task-group-spec 'asc 'creation-date #f)
              (task-group-spec 'asc 'completed? #f))
        '()
        (list (tg "2021-12-11"
                  (list (tg "Not completed" (tsl "2021-12-11 Incomplete task 1"
                                                 "2021-12-11 Incomplete task 3"))
                        (tg "Completed"     (tsl "x 2021-12-11 Complete task 2"))))
              (tg "2021-12-12"
                  (list (tg "Not completed" (tsl "2021-12-12 Incomplete task 2"))
                        (tg "Completed"     (tsl "x 2021-12-12 Complete task 1"))))
              (tg "No creation date"
                  (list (tg "Not completed" (tsl "Task without creation date")))))))
    (test-case "Nested grouping with descending completion"
      (check-grouping
        (tsl "2021-12-11 Incomplete task 1"
             "x 2021-12-12 Complete task 1"
             "Task without creation date"
             "2021-12-12 Incomplete task 2"
             "2021-12-11 Incomplete task 3"
             "x 2021-12-11 Complete task 2")
        (list (task-group-spec 'asc 'creation-date #f)
              (task-group-spec 'desc 'completed? #f))
        '()
        (list (tg "2021-12-11"
                  (list (tg "Completed"     (tsl "x 2021-12-11 Complete task 2"))
                        (tg "Not completed" (tsl "2021-12-11 Incomplete task 1"
                                                 "2021-12-11 Incomplete task 3"))))
              (tg "2021-12-12"
                  (list (tg "Completed"     (tsl "x 2021-12-12 Complete task 1"))
                        (tg "Not completed" (tsl "2021-12-12 Incomplete task 2"))))
              (tg "No creation date"
                  (list (tg "Not completed" (tsl "Task without creation date")))))))
    (test-case "Nested grouping with tags"
      (check-grouping
        (tsl "2021-12-11 Incomplete task 1 pri:B"
             "x 2021-12-12 Complete task 1 due:2022-01-02"
             "Task without creation date due:2022-01-03 pri:C"
             "2021-12-12 Incomplete task 2"
             "2021-12-11 Incomplete task 3 due:2022-01-02"
             "x 2021-12-11 Complete task 2")
        (list (task-group-spec 'desc 'tags "due")
              (task-group-spec 'asc 'creation-date #f))
        '()
        (list (tg "due:2022-01-03"
                  (list (tg "No creation date"
                            (tsl "Task without creation date due:2022-01-03 pri:C"))))
              (tg "due:2022-01-02"
                  (list (tg "2021-12-11" (tsl "2021-12-11 Incomplete task 3 due:2022-01-02"))
                        (tg "2021-12-12" (tsl "x 2021-12-12 Complete task 1 due:2022-01-02"))))
              (tg "No such tag"
                  (list (tg "2021-12-11" (tsl "2021-12-11 Incomplete task 1 pri:B"
                                              "x 2021-12-11 Complete task 2"))
                        (tg "2021-12-12" (tsl "2021-12-12 Incomplete task 2")))))))
    (test-case "Nested grouping with sorting"
      (check-grouping
        (tsl "2021-12-11 Incomplete task 1"
             "x 2021-12-12 Complete task 1"
             "Task without creation date"
             "2021-12-12 Incomplete task 2"
             "2021-12-11 Incomplete task 3"
             "x 2021-12-11 Complete task 2")
        (list (task-group-spec 'asc 'creation-date #f)
              (task-group-spec 'asc 'completed? #f))
        (list (task-sort-spec 'desc 'description))
        (list (tg "2021-12-11"
                  (list (tg "Not completed" (tsl "2021-12-11 Incomplete task 3"
                                                 "2021-12-11 Incomplete task 1"))
                        (tg "Completed"     (tsl "x 2021-12-11 Complete task 2"))))
              (tg "2021-12-12"
                  (list (tg "Not completed" (tsl "2021-12-12 Incomplete task 2"))
                        (tg "Completed"     (tsl "x 2021-12-12 Complete task 1"))))
              (tg "No creation date"
                  (list (tg "Not completed" (tsl "Task without creation date")))))))
  ))

  ; Short for "string lines"
  (define (sl . lines) (string-join lines "\n"))

  (define-check (check-grouping-result->string grouping-result
                                               omit-fields
                                               expected-result-string)
    (define result-string (group-tasks-result->string grouping-result
                                                      #:omit-fields omit-fields))
    (check string=? result-string expected-result-string))

  (run-tests
    (test-suite "Convert grouping result to multiline string"
      (test-case "Empty list"
        (check-grouping-result->string
          '()
          '()
          ""))
      (test-case "Task list"
        (check-grouping-result->string
          (tsl "2021-03-02 Task 1"
               "2021-03-05 Task 2"
               "2021-04-03 Task 3")
          '()
          (sl  "2021-03-02 Task 1"
               "2021-03-05 Task 2"
               "2021-04-03 Task 3")))
      (test-case "List with one task group"
        (check-grouping-result->string
          (list (tg "Group title" (tsl "2021-03-02 Task 1"
                                       "2021-03-05 Task 2"
                                       "2021-04-03 Task 3")))
          '()
          (sl "Group title"
              "  2021-03-02 Task 1"
              "  2021-03-05 Task 2"
              "  2021-04-03 Task 3")))
      (test-case "List with several task groups"
        (check-grouping-result->string
          (list (tg "Group 1 title" (tsl "2021-03-02 Task 1.1"
                                         "2021-03-05 Task 1.2"
                                         "2021-04-03 Task 1.3"))
                (tg "Group 2 title" (tsl "2021-03-02 Task 2.1"
                                         "2021-03-05 Task 2.2"
                                         "2021-04-03 Task 2.3"))
                (tg "Group 3 title" (tsl "2021-03-02 Task 3.1"
                                         "2021-03-05 Task 3.2"
                                         "2021-04-03 Task 3.3")))
          '()
          (sl "Group 1 title"
              "  2021-03-02 Task 1.1"
              "  2021-03-05 Task 1.2"
              "  2021-04-03 Task 1.3"
              "Group 2 title"
              "  2021-03-02 Task 2.1"
              "  2021-03-05 Task 2.2"
              "  2021-04-03 Task 2.3"
              "Group 3 title"
              "  2021-03-02 Task 3.1"
              "  2021-03-05 Task 3.2"
              "  2021-04-03 Task 3.3")))
      (test-case "List with several task groups and `omit-fields`"
        (check-grouping-result->string
          (list (tg "Group 1 title" (tsl "2021-03-02 Task 1.1"
                                         "2021-03-05 Task 1.2"
                                         "2021-04-03 Task 1.3"))
                (tg "Group 2 title" (tsl "2021-03-02 Task 2.1"
                                         "2021-03-05 Task 2.2"
                                         "2021-04-03 Task 2.3")))
          '(creation-date completion-date)
          (sl "Group 1 title"
              "  Task 1.1"
              "  Task 1.2"
              "  Task 1.3"
              "Group 2 title"
              "  Task 2.1"
              "  Task 2.2"
              "  Task 2.3")))
      (test-case "Nested groups"
        (check-grouping-result->string
          (list (tg "Group 1 title"
                  (list (tg "Group 1.1 title"
                          (tsl "2021-03-02 Task 1.1.1"
                               "2021-03-05 Task 1.1.2"
                               "2021-04-03 Task 1.1.3"))
                        (tg "Group 1.2 title"
                          (tsl "2021-03-02 Task 1.2.1"
                               "2021-03-05 Task 1.2.2"
                               "2021-04-03 Task 1.2.3"))))
                (tg "Group 2 title"
                  (list (tg "Group 2.1 title"
                          (tsl "2021-03-02 Task 2.1.1"
                               "2021-03-05 Task 2.1.2"
                               "2021-04-03 Task 2.1.3"))
                        (tg "Group 2.2 title"
                          (tsl "2021-03-02 Task 2.2.1"
                               "2021-03-05 Task 2.2.2"
                               "2021-04-03 Task 2.2.3")))))
          '()
          (sl "Group 1 title"
              "  Group 1.1 title"
              "    2021-03-02 Task 1.1.1"
              "    2021-03-05 Task 1.1.2"
              "    2021-04-03 Task 1.1.3"
              "  Group 1.2 title"
              "    2021-03-02 Task 1.2.1"
              "    2021-03-05 Task 1.2.2"
              "    2021-04-03 Task 1.2.3"
              "Group 2 title"
              "  Group 2.1 title"
              "    2021-03-02 Task 2.1.1"
              "    2021-03-05 Task 2.1.2"
              "    2021-04-03 Task 2.1.3"
              "  Group 2.2 title"
              "    2021-03-02 Task 2.2.1"
              "    2021-03-05 Task 2.2.2"
              "    2021-04-03 Task 2.2.3")))
  ))
)
