#lang racket/base

; Parser for the `todo.txt` format.
; See https://github.com/todotxt/todo.txt .

(require racket/contract)

(provide
  (contract-out
    [make-task (->* ()
                    (#:completed? boolean?
                     #:priority string-field/c
                     #:completion-date string-field/c
                     #:creation-date string-field/c
                     #:description string?
                     #:projects (listof string?)
                     #:contexts (listof string?)
                     #:tags (listof (cons/c string? string?)))
                    task?)]
    [sort-tasks (-> (listof task?) (listof task-sort-spec?) (listof task?))]
    [string->task (-> string? task?)]
    [string->tasks (-> string? (listof task?))]
    [task->string (->* (task?)
                       (#:omit-fields valid-field-omissions?)
                       string?)]
    [task-field (-> task? (or/c task-field/c task-sort-spec?) any/c)]
    [tasks->string (->* ((listof task?))
                        (#:omit-fields valid-field-omissions?)
                        string?)]
    [value-sort-func (-> task-sort-spec? (-> any/c any/c boolean?))]
    ; Although the `task-sort-spec` supertype only needs a subset of `task-field/c`,
    ; we can't use it here because the derived type `task-group-spec` needs all
    ; symbols from `task-field/c`.
    (struct task-sort-spec ([order sort-order/c]
                            [field task-field/c]))
    (struct task ([completed? boolean?]
                  [priority string-field/c]
                  [completion-date string-field/c]
                  [creation-date string-field/c]
                  [description string?]
                  [projects (listof string?)]
                  [contexts (listof string?)]
                  [tags (listof (cons/c string? string?))]))
    [valid-field-omission? (-> any/c boolean?)]
    [valid-field-omissions? (-> any/c boolean?)]
    [sort-order/c contract?]
    [task-field/c contract?])
)

(require
  racket/list
  racket/match
  racket/port
  racket/string)

; --------------------------------------------------------------------------------
; Structs and helpers

; ... -> Task
;
; A todo.txt task as described at
; https://github.com/todotxt/todo.txt#todotxt-format-rules
(struct task
  (completed?
   priority
   completion-date
   creation-date
   description
   projects
   contexts
   tags)
  #:transparent)

; We can't have a completion date without a creation date. Otherwise this would
; leave us with the completion date as single date, which is invalid in a task
; string.
(define (task-date-combination/c completion-date creation-date)
  (and (or creation-date (not completion-date)) #t))

; ... -> Task
(define (make-task
         #:completed?      [completed?      #f]
         #:priority        [priority        #f]
         #:completion-date [completion-date #f]
         #:creation-date   [creation-date   #f]
         #:description     [description     ""]
         #:projects        [projects        '()]
         #:contexts        [contexts        '()]
         #:tags            [tags            '()])
  (unless (task-date-combination/c completion-date creation-date)
    (raise-arguments-error
      'make-task
      "invalid date combination, can't have completion date without creation date"
      'completion-date completion-date
      'creation-date creation-date))
  (task
    completed?
    priority
    completion-date
    creation-date
    description
    projects
    contexts
    tags))

(define string-field/c (or/c string? #f))

; [X]  [Hash-of Task-field-symbol [Task -> X]]
;
; Needed to access task fields as determined from the command line
(define task-accessors
  (hash 'completed?      task-completed?
        'priority        task-priority
        'completion-date task-completion-date
        'creation-date   task-creation-date
        'description     task-description
        'contexts        task-contexts
        'projects        task-projects
        'tags            task-tags))

; Order-symbol Task-field-symbol -> Task-sort-spec
(struct task-sort-spec
  (; 'asc for ascending order, 'desc for descending order
   order
   ; Task field to sort by, say, 'creation-date
   field)
  #:transparent)

(define sort-order/c (or/c 'asc 'desc))
(define task-field/c (apply or/c (hash-keys task-accessors)))

; Task [Task-field-symbol | Task-sort-spec] -> Task-field-value
;
; Return field value from task according to field symbol or sort spec.
;
; Examples:
;
;   (task-field
;     (make-task #:completion-date "2021-02-11")
;     'completion-date
;   -> "2021-02-11"
;
;   (task-field
;     (make-task #:completion-date "2021-02-11")
;     (task-sort-spec 'asc 'completion-date))
;   -> "2021-02-11"
;
; The sort order in the spec doesn't influence the result.
(define (task-field task field-or-spec)
  (define accessor-func
    (if (symbol? field-or-spec)
        (hash-ref task-accessors field-or-spec)
        (hash-ref task-accessors (task-sort-spec-field field-or-spec))))
    (accessor-func task))


; --------------------------------------------------------------------------------
; Convert string to one or several tasks.

; All possibilities:
; x (A) 2011-03-02 2011-03-01 Review Tim's pull request +TodoTxt @github due:2011-03-02

; Regexp
(define task-line-regexp
  ; Don't check for invalid dates for now (e.g. month == 13).
  (let* ([date-regexp-string
          "\\d{4}-\\d{2}-\\d{2}"]
         [regexp-string
          (string-append
            "^"
            ; Completed?
            "(x )?"
            ; Priority
            "(\\([A-Z]\\) )?"
            ; Completion and/or creation date (if any)
            "((?:" date-regexp-string " ){0,2})"
            ; Description
            "(.*?)"
            ; Omit trailing whitespace.
            "\\s*"
            "$")])
    (pregexp regexp-string)))

; String String -> Integer
;
; Return number of `substr`s in `str`.
(define (count-substring str substr)
  (sub1 (length (string-split str substr))))

; String -> Task
;
; Return a `task` struct from a single-line string.
(define (string->task line)
  (define match-result
    ; We don't need the full line since we can reconstruct it from the struct
    ; fields later.
    (cdr (regexp-match task-line-regexp line)))
  ; Parse individual fields.
  ; "x" for completed task
  (define completed?
    (equal? (first match-result) "x "))
  ; For example "(B)"
  (define priority
    (if (second match-result)
        (substring (second match-result) 1 2)
        #f))
  ; ISO date values
  (define-values (completion-date creation-date)
    ; Since a matching date substring ends with a space, the number of spaces
    ; is the number of dates.
    (let* ([dates (string-split (third match-result))]
           [dates-count (length dates)])
      (match dates-count
        [0 (values #f #f)]
        [1 (values #f (first dates))]
        [2 (values (first dates) (second dates))])))
  ; Free-form text
  (define description
    (fourth match-result))
  ; "+some_project", "@some_context"
  (define-values (projects contexts)
    (let* ([parts (string-split description)]
           [collect (lambda (indicator)
                      (sort
                        (remove-duplicates
                          ; Remove indicators ("@" or "+").
                          (for/list ([part parts]
                                     #:when (string=? (substring part 0 1)
                                                      indicator))
                            (substring part 1)))
                        string<?))]
           [projects (collect "+")]
           [contexts (collect "@")])
      (values projects contexts)))
  ; Key/value pairs like "due:2021-02-21"
  (define tags
    (remove-duplicates
      (for/list ([part (string-split description)]
                #:when (equal? (count-substring part ":") 1))
        (define parts-list (string-split part ":"))
        (cons (car parts-list) (cadr parts-list)))))
  (make-task #:completed?      completed?
             #:priority        priority
             #:completion-date completion-date
             #:creation-date   creation-date
             #:description     description
             #:projects        projects
             #:contexts        contexts
             #:tags            tags))

; String -> [List-of Task]
;
; Return a list of tasks from a multiline-string. Ignore empty lines and
; those consisting of only whitespace.
(define (string->tasks str)
  (for/list ([line (port->lines (open-input-string str))]
             #:unless (string=? (string-trim line) ""))
    (string->task line)))


; --------------------------------------------------------------------------------
; Convert tasks to strings

; Helper predicates
(define valid-field-omission?
  (or/c 'completed? 'priority 'completion-date 'creation-date))

(define valid-field-omissions?
  (and/c (listof valid-field-omission?)
         (lambda (omit-fields)
           ; These aren't actual dates, but they work the same way for
           ; `task-date-combination/c`.
           (define completion-date (index-of omit-fields 'completion-date))
           (define creation-date (index-of omit-fields 'creation-date))
           ; In the context of `task->string` we're interested in omitting the
           ; fields, i.e. _not_ having them.
           (task-date-combination/c (not completion-date) (not creation-date)))))

; Task #:omit-fields [List-of Task-field-symbol] -> String
;
; Return a single-line string for `task`.
; The output should look like the original line, apart from trailing
; whitespace being removed.
;
; `#:omit-fields` can be a list of fields to omit from the task string. Only
; simple fields, i.e. 'completed?, 'priority, 'completion-date and
; 'creation-date are allowed. Omitting other data would be more complicated and
; could return odd-looking task strings (as in "Finish article for +my_project"
; -> "Finish article for").
;
; It's an error to try to omit the creation date, but not the completion date.
; This is because this would leave the completion date as the single date, but
; a single date is supposed to be a creation date.
;
; This error checking happens also if the task doesn't have a completion date,
; on the grounds that the `#:omit-fields` argument is faulty even if the string
; conversion would happen to give a correct result for the specific `task`
; argument.
(define (task->string task #:omit-fields [omit-fields '()])
  (define (may-include-field? field-symbol)
    (not (index-of omit-fields field-symbol)))
  (string-append
    (if (and (task-completed? task) (may-include-field? 'completed?))
        "x "
        "")
    (if (and (task-priority task) (may-include-field? 'priority))
        (string-append
          "("
          (task-priority task)
          ") ")
        "")
    (if (and (task-completion-date task) (may-include-field? 'completion-date))
        (string-append
          (task-completion-date task)
          " ")
        "")
    (if (and (task-creation-date task) (may-include-field? 'creation-date))
        (string-append
          (task-creation-date task)
          " ")
        "")
    (task-description task)))

; [List-of Task] #:omit-fields [List-of Task-field-symbol]-> String
;
; Return a (possibly multi-line) string for the given `task`s.
;
; See the description of `task->string` for the use of `#:omit-fields`.
(define (tasks->string tasks #:omit-fields [omit-fields '()])
  (string-join
    (for/list ([task tasks])
      (task->string task #:omit-fields omit-fields))
    "\n"))


; --------------------------------------------------------------------------------
; Sort tasks according to sort specs

; Order-symbol -> [Completed-value Completed-value -> Boolean]
;
; Return a sort function for `sort` than sorts `value1` and `value2` according
; to `sort-order`.
;
; `sort-order` is 'asc or 'desc
;
; For a `completed?` value, sort `#f` before `#t` for ascending order (i.e.
; incomplete tasks before complete tasks). For descending order use the
; opposite value order.
(define (make-cmp-completed? sort-order)
  (define (cmp-ascending value1 value2)
    (and (eq? value1 #f) (eq? value2 #t)))
  (match sort-order
    ['asc cmp-ascending]
    ; Revert sort order by swapping the (only two possible) values.
    ['desc (lambda (value1 value2)
             (cmp-ascending value2 value1))]))

; Order-symbol -> [String String -> Boolean]
;
; For other fields, always sort `#f` values after other values (i.e. put
; tasks without the field set after tasks with the field set). If the value
; isn't `#f`, i.e. the field is set, sort the values in ascending order if
; the sort spec indicates it, otherwise use descending order.
(define (make-cmp-string-field sort-order)
  (lambda (value1 value2)
    (match (list value1 value2 sort-order)
      ; Keep relative order if neither value is set.
      [(list #f #f _) #f]
      ; Sort unset values after set values, i.e. swap for values
      ; #f "some-string".
      [(list #f _ _) #f]
      ; Sort unset values after set values, i.e. don't swap for values
      ; "some-string" #f.
      [(list _ #f _) #t]
      ;
      [(list _ _ 'asc)  (string<? value1 value2)]
      [(list _ _ 'desc) (string>? value1 value2)])))

; [X]  Task-sort-spec -> [X X -> Boolean]
(define (value-sort-func sort-spec)
  (define sort-order (task-sort-spec-order sort-spec))
  (if (eq? (task-sort-spec-field sort-spec) 'completed?)
      (make-cmp-completed? sort-order)
      (make-cmp-string-field sort-order)))

; Task-sort-spec -> [Task Task -> Boolean]
;
; For a `sort-spec`, return a function for the `sort` function, so the sort
; spec can be applied to `task`s.
(define (make-task-sort-func sort-spec)
  (lambda (task1 task2)
    (define field1 (task-field task1 sort-spec))
    (define field2 (task-field task2 sort-spec))
    ((value-sort-func sort-spec) field1 field2)))

; [List-of Task] [List-of Task-sort-spec] -> [List-of Task]
;
; Sort tasks according to sort specs.
;
; Tasks are sorted according to the first sort spec, then, within each group,
; according to the second sort spec etc. In the previous sentence, "group" does
; _not_ refer to a `task-group`.
;
; Example:
;   (let ([tasks (map string->task
;                     '("2021-02-15 Clean code"
;                       "x 2021-02-18 Shopping"
;                       "Unset creation date"
;                       "x Another unset creation date"
;                       "x 2021-02-27 Code"
;                       "2020-12-01 Do laundry"))]
;         [sort-specs (list (sort-spec 'asc 'completed?)
;                           (sort-spec 'desc 'creation-date))])
;     (sort-tasks tasks sort-specs))
; results in
;   (map string->task
;        '("2021-02-15 Clean code"
;          "2020-12-01 Do laundry"
;          "Unset creation date"
;          "x 2021-02-27 Code"
;          "x 2021-02-18 Shopping"
;          "x Another unset creation date"))
(define (sort-tasks tasks sort-specs)
  (for/fold ([sorted-tasks tasks])
            ([sort-spec (reverse sort-specs)])
    (sort sorted-tasks (make-task-sort-func sort-spec))))


; --------------------------------------------------------------------------------

(module+ test
  (require
    rackunit
    al2-test-runner
    (submod ".."))

  (define-check (check-task line expected-task)
    (check-equal?
      (string->task line)
      expected-task))
  
  (run-tests
    (test-suite "Contracts"
      ; `task-date-combination/c`
      (test-case
        "Valid date combinations"
        (check-true (task-date-combination/c #f #f))
        (check-true (task-date-combination/c #f "2022-01-04"))
        (check-true (task-date-combination/c "2022-01-04" "2022-01-03")))
      (test-false
        "Completion date without creation date"
        (task-date-combination/c "2022-01-04" #f))
      ; `valid-field-omissions?`
      (test-false
        "Invalid task field symbol"
        (valid-field-omissions? '(priority invalid)))
      (test-true
        "Omit both dates"
        (valid-field-omissions? '(priority creation-date completion-date)))
      (test-true
        "Omit completion date, but keep creation date"
        (valid-field-omissions? '(priority completion-date)))
      (test-true
        "Omit neither date"
        (valid-field-omissions? '(priority)))
      (test-false
        "Omit creation date, but keep completion date"
        (valid-field-omissions? '(priority creation-date)))
  ))

  (run-tests
    (test-suite "string->task"
      ; If only a description is given, the returned task should also have only
      ; a description.
      (test-case
        "Only description"
        (check-task
          "The description"
          (make-task #:description "The description")))
      ; If the description contains trailing whitespace, it should be ignored.
      ; That is, the resulting task description shouldn't contain the trailing
      ; whitespace.
      (test-case
        "Description with trailing whitespace"
        (check-task
          "The description  "
          (make-task #:description "The description")))
      ; If a completion status and a description are in the task line, the
      ; corresponding fields should be set in the returned task. Other fields
      ; should be set to their default values.
      (test-case
        "Completion and description"
        (check-task
          "x The description"
          (make-task #:completed? #t
                     #:description "The description")))
      ; As above, but for priority and description.
      (test-case
        "Priority and description"
        (check-task
          "(B) The description"
          (make-task #:priority "B"
                     #:description "The description")))
      ; As above, but for creation date and description. (A single date should
      ; be parsed as creation date, see the specification at
      ; https://github.com/todotxt/todo.txt#todotxt-format-rules .
      (test-case
        "Creation date and description"
        (check-task
          "2021-02-13 The description"
          (make-task #:creation-date "2021-02-13"
                     #:description "The description")))
      ; If the task line contains two dates, the first should be parsed as the
      ; completion date and the second date as the creation date.
      (test-case
        "Completion date, creation date and description"
        (check-task
          "2021-02-14 2021-02-13 The description"
          (make-task #:creation-date "2021-02-13"
                     #:completion-date "2021-02-14"
                     #:description "The description")))
      ; Projects should be parsed as a list in the `projects` field.
      (test-case
        "Description with projects"
        (check-task
          "The description +project1 and +project2 (ignore another +project1 )"
          (make-task #:description "The description +project1 and +project2 (ignore another +project1 )"
                     #:projects '("project1" "project2"))))
      ; Contexts should be parsed as a list in the `contexts` field.
      (test-case
        "Description with contexts"
        (check-task
          "The description @context1 and @context2 (ignore another @context1 )"
          (make-task #:description "The description @context1 and @context2 (ignore another @context1 )"
                     #:contexts '("context1" "context2"))))
      ; Tags and their values should be parsed as a hash in the `tags` field.
      (test-case
        "Description with tags"
        (check-task
          "The description foo:bar due:2021-02-10 foo:baz (duplicate tag foo:bar )"
          (make-task #:description "The description foo:bar due:2021-02-10 foo:baz (duplicate tag foo:bar )"
                     #:tags '(("foo" . "bar") ("due" . "2021-02-10") ("foo" . "baz")))))
      ; If all fields are in the task line, they should be set accordingly in
      ; the resulting task struct.
      (test-case
        "Everything"
        (check-task
          "x (C) 2021-02-14 2021-02-13 Description with +my_project @my_context and tag:value"
          (make-task #:completed? #t
                     #:priority "C"
                     #:completion-date "2021-02-14"
                     #:creation-date "2021-02-13"
                     #:description "Description with +my_project @my_context and tag:value"
                     #:projects '("my_project")
                     #:contexts '("my_context")
                     #:tags '(("tag" . "value")))))
  ))

  (run-tests
    (test-suite "Tasks to strings"
      ; If only a description is given, the task should only contain the
      ; description. Other fields should be set to their default values.
      (test-equal?
        "Task with only description"
        (task->string (make-task #:description "The description"))
        "The description")
      ; If all fields are specified in the string, a task with the
      ; corresponding field values should be returned.
      (test-equal?
        "Task with all fields set"
        (task->string
          (make-task #:completed? #t
                     #:priority "C"
                     #:completion-date "2021-02-14"
                     #:creation-date "2021-02-13"
                     #:description "Description with +my_project @my_context and tag:value"
                     #:projects '("my_project")
                     #:contexts '("my_context")
                     #:tags '(("tag" . "value"))))
        "x (C) 2021-02-14 2021-02-13 Description with +my_project @my_context and tag:value")
      (test-equal?
        "Omit fields"
        (task->string
          (make-task #:completed? #t
                     #:priority "C"
                     #:completion-date "2021-02-14"
                     #:creation-date "2021-02-13"
                     #:description "Description with +my_project @my_context and tag:value"
                     #:projects '("my_project")
                     #:contexts '("my_context")
                     #:tags '(("tag" . "value")))
          #:omit-fields '(priority completed? creation-date completion-date))
        "Description with +my_project @my_context and tag:value")
      (test-exn
        ; Since a single date in the task must be a creation date, using the
        ; completion date as single date should be forbidden.
        "Request to omit 'creation-date, but not 'completion-date"
        exn:fail:contract?
        (lambda ()
          ; We only want to check `#:omit-fields`, the task's fields aren't
          ; relevant here.
          (task->string (task) #:omit-fields '(creation-date))))
      (test-equal?
        "List of single task to string"
        (tasks->string
          (list
            (make-task #:completed? #t
                       #:priority "C"
                       #:completion-date "2021-02-14"
                       #:creation-date "2021-02-13"
                       #:description "Description with +my_project @my_context and tag:value"
                       #:projects '("my_project")
                       #:contexts '("my_context")
                       #:tags '(("tag" . "value")))))
        "x (C) 2021-02-14 2021-02-13 Description with +my_project @my_context and tag:value")
      (test-equal?
        "List of two tasks to multi-line string"
        (tasks->string
          (list
            (make-task #:description "The description")
            (make-task #:completed? #t
                       #:creation-date "2021-02-13"
                       #:description "Description with +my_project @my_context and tag:value"
                       #:projects '("my_project")
                       #:contexts '("my_context")
                       #:tags '(("tag" . "value")))))
        "The description\nx 2021-02-13 Description with +my_project @my_context and tag:value")
      (test-equal?
        "Omit fields in list of tasks"
        (tasks->string
          (list
            (make-task #:description "The description")
            (make-task #:completed? #t
                       #:creation-date "2021-02-13"
                       #:description "Description with +my_project @my_context and tag:value"
                       #:projects '("my_project")
                       #:contexts '("my_context")
                       #:tags '(("tag" . "value"))))
          ; If we omit the creation date, we must also omit the completion date.
          #:omit-fields '(completion-date creation-date))
        "The description\nx Description with +my_project @my_context and tag:value")
  ))

  (run-tests
    (test-suite "string->tasks"
      ; If an empty string is given, an empty list should be returned as there
      ; are no tasks encoded in the string.
      (test-equal?
        "Empty string"
        (string->tasks "")
        '())
      ; Lines consisting of only whitespace should be ignored.
      (test-equal?
        "Several whitespace-only lines"
        (string->tasks "\n  \n \n\n  \n")
        '())
      ; If multiple task lines are given, the corresponding task values should
      ; be in the resulting list in the order of the corresponding strings.
      ; Lines consisting of only whitespace should be ignored.
      (test-equal?
        "Multiple tasks"
        (string->tasks "\nx First task \n  \n(A) Second task\n\n")
        (list (make-task #:completed? #t
                         #:description "First task")
              (make-task #:priority "A"
                         #:description "Second task")))
  ))

  ; Compare the calculated list of tasks strings with the expected list of
  ; task strings.
  (define-check (check-sorting
                 before-task-strings
                 sort-specs
                 expected-after-task-strings)
    (define tasks (map string->task before-task-strings))
    (define sorted-tasks (sort-tasks tasks sort-specs))
    (define after-task-strings (map task->string sorted-tasks))
    (check-equal? after-task-strings expected-after-task-strings))

  (run-tests
    (test-suite "Sorting"
      ; If no tasks are given, an empty list (i.e. no tasks) should be
      ; returned.
      (test-case "No tasks"
        (check-sorting '()
                       (list (task-sort-spec 'desc 'completion-date))
                       '()))
      ; If a single task is given, a list with this task should be returned.
      ; (There's nothing to sort.)
      (test-case "Single task"
        (check-sorting '("x 2021-02-21 Description with +project")
                       (list (task-sort-spec 'asc 'creation-date))
                       '("x 2021-02-21 Description with +project")))
      ; If no sort specs are given, the original task list should be returned.
      (test-case "No sort specs"
        (check-sorting '("x 2021-02-21 Description with +project"
                         "Shopping")
                       '()
                       '("x 2021-02-21 Description with +project"
                         "Shopping")))
      ; If sorting by ascending completion status is requested, sort completed
      ; tasks _after_ incomplete tasks. Otherwise keep the relative order. Think
      ; of a "true" completed flag as 1 and a "false" flag as 0.
      (test-case "Ascending completion"
        (check-sorting '("x Do laundry"
                         "Shopping"
                         "Incomplete"
                         "x Clean code")
                       (list (task-sort-spec 'asc 'completed?))
                       '("Shopping"
                         "Incomplete"
                         "x Do laundry"
                         "x Clean code")))
      ; If sorting by descending completion status is requested, sort completed
      ; tasks _before_ incomplete tasks. Otherwise keep the relative order.
      (test-case "Descending completion"
        (check-sorting '("x Do laundry"
                         "Shopping"
                         "Incomplete"
                         "x Clean code")
                       (list (task-sort-spec 'desc 'completed?))
                       '("x Do laundry"
                         "x Clean code"
                         "Shopping"
                         "Incomplete")))
      ; If sorting by ascending creation date is requested, tasks should be
      ; sorted by ascending creation date.
      (test-case "Ascending creation date"
        (check-sorting '("2021-02-15 Clean code"
                         "2021-02-18 Shopping"
                         "2020-12-01 Do laundry")
                       (list (task-sort-spec 'asc 'creation-date))
                       '("2020-12-01 Do laundry"
                         "2021-02-15 Clean code"
                         "2021-02-18 Shopping")))
      ; If sorting by descending creation date is requested, tasks should be
      ; sorted by descending creation date.
      (test-case "Descending creation date"
        (check-sorting '("2021-02-15 Clean code"
                         "2021-02-18 Shopping"
                         "2020-12-01 Do laundry")
                       (list (task-sort-spec 'desc 'creation-date))
                       '("2021-02-18 Shopping"
                         "2021-02-15 Clean code"
                         "2020-12-01 Do laundry")))
      ; If sorting by ascending creation date is requested, tasks should be
      ; sorted by ascending creation date. Tasks without creation date should
      ; be sorted after tasks with creation date, keeping their relative order.
      (test-case "Ascending creation date with missing date"
        (check-sorting '("2021-02-15 Clean code"
                         "Unset creation date"
                         "2021-02-18 Shopping"
                         "Another unset creation date"
                         "2020-12-01 Do laundry")
                       (list (task-sort-spec 'asc 'creation-date))
                       '("2020-12-01 Do laundry"
                         "2021-02-15 Clean code"
                         "2021-02-18 Shopping"
                         "Unset creation date"
                         "Another unset creation date")))
      ; If sorting by descending creation date is requested, tasks should be
      ; sorted by descending creation date. Tasks without creation date should
      ; be sorted after tasks with creation date, keeping their relative order.
      (test-case "Descending creation date with missing date"
        (check-sorting '("2021-02-15 Clean code"
                         "Unset creation date"
                         "2021-02-18 Shopping"
                         "Another unset creation date"
                         "2020-12-01 Do laundry")
                       (list (task-sort-spec 'desc 'creation-date))
                       '("2021-02-18 Shopping"
                         "2021-02-15 Clean code"
                         "2020-12-01 Do laundry"
                         "Unset creation date"
                         "Another unset creation date")))
      ; If multiple sort specs are given, they should determine the "sort
      ; groups" from outside to inside. For example, sorting first according to
      ; the first sort spec, then in case of ties, according to the second sort
      ; spec and so on.
      (test-case "Sort by completion, then creation date"
        (check-sorting '("2021-02-15 Clean code"
                         "x 2021-02-18 Shopping"
                         "Unset creation date"
                         "x Another unset creation date"
                         "x 2021-02-27 Code"
                         "2020-12-01 Do laundry")
                       (list (task-sort-spec 'asc 'completed?)
                             (task-sort-spec 'desc 'creation-date))
                       '("2021-02-15 Clean code"
                         "2020-12-01 Do laundry"
                         "Unset creation date"
                         "x 2021-02-27 Code"
                         "x 2021-02-18 Shopping"
                         "x Another unset creation date")))
  ))
)
