#lang scribble/manual

@(let ([this-handler (scribble-eval-handler)])
   (scribble-eval-handler
     (λ (f show? expr)
       (with-handlers ([exn:fail? values])
         (this-handler f show? expr)))))

@(require
   scribble/example
   (for-label racket)
   (for-label file/todo-txt))

@(define helper-eval (make-base-eval))
@examples[
  #:eval helper-eval
  #:hidden
  (require racket/pretty
           racket/string
           file/todo-txt)]

@(define todo.txt (hyperlink "https://github.com/todotxt/todo.txt" "todo.txt"))


@title{Processing of todo.txt tasks}
@author{Stefan Schwarzer}

This collection provides APIs to process tasks from @|todo.txt|-formatted
files:
@itemlist[
  @item{Convert strings to @racket[task] values}
  @item{Convert @racket[task] values to strings}
  @item{Group and sort @racket[task] values, also hierarchically}
  @item{Convert group and sort results to strings}]

For the command-line interface, @code{todoreport}, see the
@(hyperlink "https://git.sr.ht/~sschwarzer/todo-txt/tree/main/item/README.md"
@code{README.md}) file.

@defmodule[file/todo-txt]

@; --------------------------------------------------------------------------------

@section{Tasks}

@; --------------------------------------------------

@subsection{Creating @racket[task]s}

@defstruct*[task ([completed? boolean?]
                  [priority string-field/c]
                  [completion-date string-field/c]
                  [creation-date string-field/c]
                  [description string?]
                  [projects (listof string?)]
                  [contexts (listof string?)]
                  [tags (listof (cons/c string? string?))])
                 #:transparent]{
  A @todo.txt task, where @racket[string-field/c] is @racket[(or/c string? #f)].}

However, usually it will be easier to use the following helper function
@racket[make-task].

@defproc[(make-task
          [#:completed? completed? boolean? #f]
          [#:priority priority string-field/c #f]
          [#:completion-date completion-date string-field/c #f]
          [#:creation-date creation-date string-field/c #f]
          [#:description description string? ""]
          [#:projects projects (listof string?) '()]
          [#:contexts contexts (listof string?) '()]
          [#:tags tags (listof (cons/c string? string?)) '()])
         task?]{
  Converts zero or more keyword arguments to a @racket[task].

  Note that @racket[make-task] prevents the creation of a @racket[task] with a
  completion date, but no creation date. This is because it's not possible to
  express such a task in the @todo.txt string form, since a single date is
  supposed to be creation date, not a completion date. If it turns out that
  there are applications where you need to create the task with a completion
  date only and add the creation date later, before converting the task to a
  string, the restriction in @racket[make-task] may be removed.}

Tasks can also be parsed from strings:

@defproc[(string->task
          [line string?])
         task?]{
  Converts a single-line string to a @racket[task]. Trailing whitespace is removed from
  the description.

  @examples[
    #:eval helper-eval
    (string->task "x 2021-04-02 Task @home +todo.txt due:2021-04-03 ")]

  Note that a single date is parsed as a creation date.}

@defproc[(string->tasks
          [str string?])
         (listof task?)]{
  Converts a multi-line string to a list of tasks. This function ignores lines that
  are empty or contain only whitespace.

  @examples[
    #:eval helper-eval
    (string->tasks "x A completed task @home\n \nAn incomplete task @work")]}

@; --------------------------------------------------

@subsection{Converting @racket[task]s to strings}

@defproc[(task->string
          [task task?]
          [#:omit-fields omit-fields (listof symbol?) '()])
         string?]{
  Converts a single task to a single-line string. The allowed symbols for
  @racket[_omit-fields] are @racket['completed?], @racket['priority],
  @racket['completion-date] and @racket['creation-date].

  There's a similar restriction as in @racket[make-task]: If you include
  @racket['creation-date] in the @racket[_omit-fields] list, you must also
  include @racket['completion-date], because you can't get a valid string
  representation for the task if you have a single date which is a completion
  date. Note that this restriction is also applied if the concrete
  @racket[_task] doesn't have a completion date. The rationale is that if you
  call @racket[task->string] on different tasks with the same
  @racket[_omit-fields] argument, all of the conversions should succeed or all
  the conversions should fail.

  @examples[
    #:eval helper-eval
    (define task (make-task #:completed? #t
                            #:creation-date "2021-04-02"
                            #:description "Uninteresting task"))
    (task->string task)
    (task->string task
                  #:omit-fields '(creation-date completion-date))]}

@defproc[(tasks->string
          [tasks (listof task?)]
          [#:omit-fields omit-fields (listof symbol?) '()])
         string?]{
  Converts a list of tasks to a (possibly multi-line) string.}

@; --------------------------------------------------

@subsection{Sorting @racket[task]s}

To sort a list of @racket[task]s, you must provide a list of
@racket[task-sort-spec]s:

@defstruct*[task-sort-spec ([order (or/c 'asc 'desc)]
                            [field symbol?])
                           #:transparent]{
  The argument @var[order] is one of the symbols @racket['asc] or
  @racket['desc] to sort in ascending or descending order, respectively.

  The @var[field] argument is one of the symbols @racket['completed?],
  @racket['priority] etc., which correspond to the @racket[task] fields. Since
  tasks can have more than one context, project or tag, respectively, it
  doesn't make sense to sort by context, for example. Therefore
  @racket['contexts] etc. aren't valid @var[field] values.

  For a @racket['completed?] field, ascending order means sorting incomplete
  tasks before completed tasks.}

@defproc[(sort-tasks
          [tasks (listof task?)]
          [sort-specs (listof task-sort-spec?)])
         (listof task?)]{
  Sort @racket[task]s according to @racket[sort-spec]s and return a new list
  containing the sorted @racket[task]s.

  Tasks that don't contain the field to sort by are moved to the end of the
  sorted list -- regardless of ascending or descending order --, keeping their
  original relative order. Note that tasks always have a @racket['completed?]
  field, which is @racket[#f] for incomplete tasks.

  If the @var[sort-specs] list contains more than one @racket[task-sort-spec],
  the tasks are sorted according to the first @racket[task-sort-spec] first,
  then sets of tasks that sorted equally, are sorted by the second
  @racket[task-sort-spec], and so on.

  The sort is stable, so if tasks are sorted by a given field and have the same
  field value, the tasks remain in their previous relative order.

  If @var[sort-specs] is an empty list, the original list of tasks is returned
  unchanged.

  Here are a few examples.

  @examples[
    #:eval helper-eval
    #:label "First let's define a few tasks to sort:"
    (define my-tasks
      (string->tasks
        (string-join
          '("x 2021-04-02 Early complete task"
            "2021-04-03 Late incomplete task"
            "2021-04-02 Early incomplete task"
            "x 2021-04-03 Late complete task")
          "\n")))
    my-tasks]

  @examples[
    #:eval helper-eval
    #:label "Sort the tasks by a single sort spec:"
    (displayln
      (tasks->string
        (sort-tasks my-tasks (list (task-sort-spec 'asc 'completed?)))))]

  @examples[
    #:eval helper-eval
    #:label "Use descending order to sort completed tasks first:"
    (displayln
      (tasks->string
        (sort-tasks my-tasks (list (task-sort-spec 'desc 'completed?)))))]

  Sort by @racket['completed?] , then @racket['creation-date] :
  @examples[
    #:eval helper-eval
    #:label #f
    (displayln
      (tasks->string
        (sort-tasks my-tasks (list (task-sort-spec 'asc 'completed?)
                                   (task-sort-spec 'asc 'creation-date)))))]

  Sort by @racket['creation-date] , then @racket['completed?] :
  @examples[
    #:eval helper-eval
    #:label #f
    (displayln
      (tasks->string
        (sort-tasks my-tasks (list (task-sort-spec 'asc 'creation-date)
                                   (task-sort-spec 'asc 'completed?)))))]

  Tasks without the field are always sorted last, regardless of ascending or
  descending sort order:

  @examples[
    #:eval helper-eval
    #:label #f
    (define my-tasks
      (string->tasks
        (string-join
          '("2021-04-02 Early task"
            "Task without creation date"
            "2021-04-03 Late task")
          "\n")))
    my-tasks
    (displayln
      (tasks->string
        (sort-tasks my-tasks (list (task-sort-spec 'asc 'creation-date)))))
    (displayln
      (tasks->string
        (sort-tasks my-tasks (list (task-sort-spec 'desc 'creation-date)))))]}

@; --------------------------------------------------------------------------------

@section{Task groups}

Tasks can't only be sorted, but also be grouped. Both operations are similar,
but not the same:

@itemlist[
  @item{Sorting returns a flat list of tasks. On the other hand, grouping
    returns a nested structure, a list of @racket[task-group]s, where each
    group contains a title and a list of @racket[task]s or @racket[task-group]s.}
  @item{A task can be included in more than one group. This happens when grouping
    by a task field that can occur several times in a single task, namely the
    fields @racket['contexts], @racket['projects] and @racket['tags]. For
    example, the task @racket["Example task @work @phone"], when grouping by
    contexts, will be included in two groups, one with the title
    @racket["@phone"] and one with the title @racket["@work"]. There's also a
    way to group by a given tag key.}]

@; --------------------------------------------------

@subsection{Creating @racket[task-group]s}

@defstruct*[task-group ([title string?]
                        [contents (or/c (listof task-group?) (listof task?))])
                       #:transparent]{
  A group of @racket[task]s, possibly nested in other @racket[task-group]s. You
  shouldn't create @racket[task-group] values yourself; they're created by
  @racket[group-tasks].}

@; --------------------------------------------------

@subsection{Grouping @racket[task]s}

In order to group tasks, we need @racket[task-group-spec] structs:

@defstruct*[(task-group-spec task-sort-spec) ([tag-key (or/c string? #f)])
            #:transparent]{
  A struct to specify how a list of @racket[task]s should be grouped.

  A @racket[task-group-spec] is similar to a @racket[task-sort-spec], but with
  the following differences:
  @itemlist[
    @item{The @racket[task-sort-spec] @racket[_field] field can also contain
      @racket['contexts], @racket['projects] and @racket['tags], which don't
      make sense for sorting, only for grouping.}
    @item{To group by tag values, specify the @racketid[tag-key] field as a
      string for the tag. For example, to group tasks by due date, use
      @racket[(task-group-spec 'asc 'tags "due")]. When not grouping by a tag,
      set @racket[tag-key] to @racket[#f].}]}

With @racket[task-group-spec] explained, we can group tasks:

@defproc[(group-tasks [tasks (listof task?)]
                      [group-specs (listof task-group-spec?) '()]
                      [sort-specs (listof task-sort-spec?) '()])
         (or/c (listof task-group?) (listof task?))]{

  Groups @racket[tasks] according to @racket[task-group-specs].

  If @racket[sort-specs] isn't an empty list, all task lists inside the groups
  are sorted according to @racket[sort-specs]. This argument behaves the same
  as for @racket[sort-tasks].

  If there are tasks that don't have a "natural" group (example: you group by
  @racket['contexts] and a task doesn't have any contexts),
  @racket[group-tasks] appends a @racket[task-group] with an appropriate
  name as the last group, say, @racket["No contexts"]. This special group
  @emph{always} comes at the end, regardless of whether the group order is
  specified as @racket['asc] or @racket['desc].

  @examples[
    #:eval helper-eval
    #:hidden
    (pretty-print-columns 65)]

  @examples[
    #:eval helper-eval
    #:label "Here's an example with ascending order:"
    (define my-tasks
      (string->tasks
        (string-join
          '("Task @home"
            "Task without context"
            "Task @work")
          "\n")))
    my-tasks
    (group-tasks my-tasks (list (task-group-spec 'asc 'contexts #f)))]

  Note that the group with the title @racket["@home"] is sorted alphabetically
  before the group with the title @racket["@work"], whereas the title
  @racket["No contexts"] appears last.

  @examples[
    #:eval helper-eval
    #:label "Here's the same example, but with descending order:"
    (group-tasks my-tasks (list (task-group-spec 'desc 'contexts #f)))]

  Note that the order of the groups titled @racket["@home"] and
  @racket["@work"] is swapped now, but the group titled @racket["No contexts"]
  still appears last.

  Note that @racket['completed?] in a @racket[task-group-spec] @emph{seems} to
  behave differently. You can control whether non-completed tasks come before
  completed tasks or the other way around by specifying @racket['asc] or
  @racket['desc] group order, respectively. This is because @emph{every} task
  has a @racket[completed?] field, which has the value @racket[#f] for
  non-completed tasks. Therefore, the group title @racket["Not completed"] is
  used for tasks with a @racket[completed?] value of @racket[#f] instead of
  indicating a ``missing'' completion status.

  @examples[
    #:eval helper-eval
    #:label "See the difference in this example:"
    (define my-tasks
      (string->tasks
        (string-join
          '("x Complete task"
            "Incomplete task")
          "\n")))
    my-tasks
    (group-tasks my-tasks (list (task-group-spec 'asc 'completed? #f)))
    (group-tasks my-tasks (list (task-group-spec 'desc 'completed? #f)))]

  Note how the order of the groups changes.

  Nested grouping is supported. Each additional @racket[task-group-spec] in
  the @racket[_group-specs] argument causes an additional level of
  @racket[task-group]s.

  @examples[
    #:eval helper-eval
    #:hidden
    (pretty-print-columns 80)]

  @examples[
    #:eval helper-eval
    #:label "For example, we can group by completion status and for each
      completion status group by priority:"
    (define my-tasks
      (string->tasks
        (string-join
          '("x (B) Complete task with priority B"
            "(A) Incomplete task with priority A"
            "(C) Incomplete task with priority C"
            "(C) Other incomplete task with priority C"
            "x (A) Complete task with priority A")
          "\n")))
    my-tasks
    (group-tasks my-tasks
                 (list (task-group-spec 'asc 'completed? #f)
                       (task-group-spec 'asc 'priority #f)))]

  Similarly, it's possible to specify an empty list for the
  @racket[_group-specs] argument. In this case, the result doesn't contain any
  @racket[task-group]s, that is, the behavior is the same as for
  @racket[sort-tasks].}

@; --------------------------------------------------

@subsection{Converting a @racket[group-tasks] result to a string}

@defproc[(group-tasks-result->string [group-tasks-result
                                      (or/c (listof task-group?) (listof task?))]
                                     [#:omit-fields omit-fields (listof symbol?) '()])
         string?]{
  Converts the result of a @racket[group-tasks] result to a string. For the
  description of the @racket[_omit-fields] argument, refer to
  @racket[task->string].

  @examples[
    #:eval helper-eval
    #:label "This example shows string representations of the previous nested
      grouping result:"
    (define my-tasks
      (string->tasks
        (string-join
          '("x (B) Complete task with priority B"
            "(A) Incomplete task with priority A"
            "(C) Incomplete task with priority C"
            "(C) Other incomplete task with priority C"
            "x (A) Complete task with priority A")
          "\n")))
    my-tasks
    (define my-grouping-result
      (group-tasks my-tasks
                   (list (task-group-spec 'asc 'completed? #f)
                         (task-group-spec 'asc 'priority #f))))
    (displayln (group-tasks-result->string my-grouping-result))
    (displayln (group-tasks-result->string
                 my-grouping-result
                 #:omit-fields '(completed? priority)))]}

@; --------------------------------------------------------------------------------

@section{Questions and answers}

@bold{How can I change the order of completed and non-completed tasks?}

To sort non-completed tasks before completed tasks, use a
@racket[task-sort-spec] with order @racket['asc]. To sort completed tasks
before non-completed tasks, use the order @racket['desc]. The same applies to
the @racket[_group-specs] argument when calling @racket[group-tasks].

@bold{How can I group by tag values?}

In the @racket[_group-specs], use a @racket[task-group-spec] with the third
field specifying the name of the tag. For example, use @racket[(task-group-spec
'asc 'tags "due")] to group by ascending due date.

@bold{How can I sort tasks inside groups?}

By default, the relative order of tasks in each group is the same as in the
original task list passed to @racket[group-tasks]. To influence the order
of tasks inside groups, pass a @racket[_sort-specs] argument as the third
argument to @racket[group-tasks].

@examples[
  #:eval helper-eval
  #:label "For example, to group by priority and sort the tasks in each group
    by their descriptions, use:"
  #:no-result
  (group-tasks my-tasks
               (list (task-group-spec 'asc 'priority #f))
               (list (task-sort-spec 'asc 'description)))]

@bold{Is the result of @racket[group-tasks] always a list of @racket[task-group]s?}

Usually yes, but not always. If the @racket[_group-specs] argument for
@racket[group-tasks] is an empty list, there's nothing to group by and the
result is a list of @racket[task]s, not a list of @racket[task-group]s.

@bold{Is it an error to pass an empty task list to @racket[sort-tasks] or
  @racket[group-tasks]?}

Although passing an empty task list @emph{may} come from a bug in your code,
from the point of view of @racket[sort-tasks] and @racket[group-tasks], it's
not an error to pass an empty task list. For an empty list, the result of either
function will be an empty list.

Background: Sorting a list of tasks gives a list with the same tasks, but
sorted. So sorting an empty task list returns an empty task list. Similarly,
when grouping a list of tasks, groups without tasks are omitted. Therefore,
grouping an empty task list results in only empty groups, and since they're
omitted, the overall result is an empty list.

@bold{Is it an error to pass an empty task list to @racket[tasks->string] or
  @racket[group-tasks-result->string]?}

Similarly to the previous question, it's not an error to pass an empty tasks
list. If you do, the result will be an empty string.

Background: Converting a list of tasks to a string gives a string with as many
lines as tasks, with one line per task. So a list with zero tasks is converted
to a string with zero lines, i.e. an empty string. The same logic applies to
grouping, considering that empty groups are removed from the output (see the
previous question).
