#lang racket/base

(require racket/contract)

(provide
  (contract-out
    [make-command-line-data (-> command-line-data?)]
    (struct command-line-data ([tasks (listof task?)]
                               [group-specs (listof task-group-spec?)]
                               [sort-specs (listof task-sort-spec?)]
                               [field-omission-specs valid-field-omissions?]))))

(require
  racket/cmdline
  racket/file
  racket/list
  racket/match
  racket/string
  "../../todo-txt.rkt")


; --------------------------------------------------------------------------------
; Command line processing

(define PROGRAM "todoreport")

; -> Command-line-data
(struct command-line-data
  (; [List-of Task]
   tasks
   ; [List-of Task-group-spec]
   group-specs
   ; [List-of Task-sort-spec]
   sort-specs
   ; [List-of Symbol] (only certain symbols are allowed)
   field-omission-specs)
  #:transparent)

; This is the same environment variable as used by the `todo.sh` script, see
; https://github.com/todotxt/todo.txt-cli/blob/master/todo.cfg
(define TODO-FILE-ENV-NAME "TODO_FILE")

; --------------------------------------------------
; Spec parsing and processing

; Specs are a way for the user to specify group and sort options, as well as
; fields to omit in the output, on the command line.
;
; There are three kinds of specs:
;
; - Sort specs: These are instances of the `task-sort-spec` struct.
;
;   The sort spec strings to specify on the command line have the form "x",
;   "prio" (or "prio"), "cod", "crd" and "desc" for completion status,
;   priority, completion date, creation date and description, respectively.
;
;   Each of the strings can optionally be prefixed with an order indicator
;   "+" for ascending order (the default) or "-" for descending order. For
;   example, "-crd" means sorting by descending creation date.
;
;   Several sort specs can be specified by separating them with commas.
;   For example, "x,-crd" means: sort by completion status, then sort
;   tasks with each completion status setting by descending creation date.
;
;   By the way, an ascending sort for the completion status means sorting
;   incomplete tasks first.
;
;   If a task doesn't have a given field you want to sort by, for example, you
;   sort by creation date, but some tasks don't have a creation date, those are
;   sorted after tasks with creation date. This applies regardless of ascending
;   or descending sort order.
;
; - Group specs: Instances of `task-group-spec`
;
;   In addition to the above task field strings (say, "x" or "crd"), spec
;   strings for grouping can be "@", "+" and "<tag-key>:" to group by contexts,
;   projects and tags with a given key, respectively. For example, "-+,-due:"
;   means grouping by project and sort the project names in descending order.
;   In each project group, group by descending due date.
;
;   Each group spec will lead to an additional nesting level in the output.
;
;   This may all sound complicated, but it's relatively simple once you do a
;   few experiments on the command line.
;
;   Similar to sorting, grouping tasks by a field they don't have, puts them in
;   a separate group which always comes after the "real" groups. For example,
;   if you group by projects, tasks without any projects will be put in a
;   group titled "No project".
;
; - Field omission specs: These are plain Racket symbols, like 'priority.
;
;   The purpose of field omission specs is to remove task fields from the
;   output that you're not interested in. For example, when grouping by
;   priority, the priorities are visible in the group titles, so it's not
;   necessary to show the priority for each task line.
;
;   A few specialties to keep in mind:
;
;   - Field spec strings for field omissions must not contain an order
;     indicator like "+" or "-" because that doesn't make sense.
;
;   - When you specify "crd" for the creation date this implies omitting the
;     completion date as well. This is because a single date in a task line
;     always means the creation date, so it would be misleading to omit the
;     creation date, leaving the completion date as single date, which would
;     then be wrongly understood as creation date.
;
;   - It's only possible to omit simple fields with "x", "pri" (or "prio"),
;     "cod" and "crd". You can't omit contexts, projects or tags. You can't
;     omit the description field, either.
;
;   - It's possible to specify the value "auto", which means to take fields
;     to omit from the group specs. In line with the previous point, only group
;     specs for simple fields will be included here, and omitting the creation
;     date will again imply omitting the completion date, too.

(define spec-hash
  (hash "x"    'completed?
        "pri"  'priority
        "prio" 'priority
        "cod"  'completion-date
        "crd"  'creation-date
        "desc" 'description
        "@"    'contexts
        "+"    'projects))

(define order-indicator-hash
  (hash "+" 'asc
        "-" 'desc
        ; Acending is default.
        ""  'asc))

; String Spec-type-symbol
; -> [Listof [Task-sort-spec | Task-group-spec | Task-field-symbol | String]]
;
; `Spec-type-symbol` can be one of 'sort, 'group or 'omit-fields.
;
; Return spec structs from a string with zero or more specs. If a spec string
; can't be parsed for the given spec type, that string is kept as a string for
; later post-processing or error reporting.
;
; Examples:
;   (specs-string->specs "x,-crd" 'sort)
;   -> (list (task-sort-spec 'asc 'completed?) (task-sort-spec 'desc 'creation-date))
;
;   (specs-string->specs "x,-+" 'group)
;   -> (list (task-group-spec 'asc 'completed? #f)
;            (task-group-spec 'desc 'projects #f))
;
;   (specs-string->specs "x,crd" 'omit-fields)
;   -> '(completed? creation-date)
;
;   ; Project fields can't be omitted.
;   (specs-string->specs "x,+" 'omit-fields)
;   -> '(completed? "+")
;
;   (specs-string->specs "auto,x" 'omit-fields)
;   -> '("auto" 'completed?)
;
; As the last example shows, the processing of the special "auto" string
; happens after parsing with `specs-string->specs`.
(define (specs-string->specs specs-string spec-type)
  (define order-indicator-regexp-string
    (case spec-type
      [(sort group)  "(\\+|\\-|)"]
      [(omit-fields) "()"]))
  (define allowed-spec-strings
    (case spec-type
      [(sort)        '("x" "pri" "prio" "cod" "crd" "desc")]
      [(group)       '("x" "pri" "prio" "cod" "crd" "desc" "@" "+")]
      [(omit-fields) '("x" "pri" "prio" "cod" "crd")]))
  (define tags-regexp-string
    (case spec-type
      [(sort omit-fields) ""]
      ; Include "|" to join with preceding regexes.
      [(group)            "|(?:\\S+:)"]))
  (define spec-regexp
    (pregexp (string-append
               "^"
               order-indicator-regexp-string
               "("
                 (string-join (map regexp-quote allowed-spec-strings) "|")
                 tags-regexp-string
               ")"
               "$")))
  (for/list ([spec-string-with-order-indicator (string-split specs-string ",")])
    (define match-result (regexp-match spec-regexp spec-string-with-order-indicator))
    (cond
      [match-result
       (match-define (list _ order-indicator spec-string) match-result)
       (case spec-type
         [(sort)
          (task-sort-spec (hash-ref order-indicator-hash order-indicator)
                          (hash-ref spec-hash spec-string))]
         [(group)
          (define tag-spec? (string-suffix? spec-string ":"))
          (task-group-spec (hash-ref order-indicator-hash order-indicator)
                           (if tag-spec?
                               'tags
                               (hash-ref spec-hash spec-string))
                           (if tag-spec?
                               ; Tag name, e.g. "due".
                               (substring spec-string 0 (sub1 (string-length spec-string)))
                               #f))]
         [(omit-fields)
          (hash-ref spec-hash spec-string)])]
      [else
       ; Keep string for post-processing.
       spec-string-with-order-indicator])))

; Apply the following functions in this order:
;
; 1. `update-omit-fields-specs` (only for field omission specs)
; 2. `check-specs`
; 3. `clean-specs`

; [Listof [Task-field-symbol | String]] -> [Listof [Task-field-symbol | String]]
;
; Change the string "auto" in a list `omit-field-specs` to the corresponding
; fields from the `group-specs`. Note that complex fields won't a considered in
; the field omission specs and that for a creation date to omit also the
; completion date will be omitted. In other words, if `omit-field-specs` after
; the "auto" conversion contains `'creation-date` but not `'completion-date`,
; the latter will be added to the result.
;
; The implementation requires that `group-specs` is valid, i.e. has already
; been checked for invalid entries (strings).
(define (update-field-omission-specs field-omission-specs group-specs)
  (define group-spec-fields (map task-sort-spec-field group-specs))
  (define may-need-completion-date
    (flatten
      (for/list ([spec field-omission-specs])
        (if (equal? spec "auto")
          (for/list ([group-spec-field group-spec-fields]
                     #:when (valid-field-omission? group-spec-field))
            group-spec-field)
          spec))))
  (define index-of-completion-date (index-of may-need-completion-date 'completion-date))
  (define index-of-creation-date   (index-of may-need-completion-date 'creation-date))
  (cond
    [(and index-of-creation-date (not index-of-completion-date))
     (define-values (head tail) (split-at may-need-completion-date index-of-creation-date))
     (append head (list 'completion-date) tail)]
    [else
     may-need-completion-date]))

; Listof [Task-sort-spec | Task-group-spec | Task-field-symbol | String]] String ->
;
; Check `specs`, which can be sort specs, group specs or field omission specs.
;
; If there's a string present, raise a user error using `specs-type-string` in
; the error message.
;
; Example:
;   ; Can't remove contexts field from tasks.
;   (check-specs '(priority "@") "field omission specs")
(define (check-specs specs specs-type-string)
  (define invalid-specs (filter string? specs))
  (when (not (null? invalid-specs))
    (raise-user-error (format "~a: invalid ~a: ~a"
                              PROGRAM
                              specs-type-string
                              (string-join invalid-specs)))))

; "Clean" and return `specs` list. Currently, this only removes duplicate list
; items.
(define (clean-specs specs)
  (remove-duplicates specs))

; --------------------------------------------------
; Task file paths

; String -> [List-of Task]
;
; Return a list of all `task`s from file `path`.
(define (file->tasks path)
  ; Use `string->tasks` instead of `string->task` to implicitly ignore empty
  ; lines.
  (append* (map string->tasks (file->lines path))))

; --------------------------------------------------
; Command line

; -> Command-line-data
;
; Parse the comamnd line and return a `command-line-data` struct value.
(define (make-command-line-data)
  (define group-specs-string "")
  (define sort-specs-string "")
  (define field-omission-specs-string "")
  (define todo-file-paths-pre #f)
  (command-line
    #:program PROGRAM
    #:once-each
    [("-g" "--group-by") specs
                         "Group specs"
                         (set! group-specs-string specs)]
    [("-s" "--sort-by") specs
                        "Sort specs"
                        (set! sort-specs-string specs)]
    [("-o" "--omit-fields") specs
                            "Field omission specs"
                            (set! field-omission-specs-string specs)]
    #:ps
      ""
      "Read todo.txt files, rearrange them and print them. See"
      "https://github.com/todotxt/todo.txt/blob/master/README.md for the todo.txt"
      "format description."
      ""
      "The \"specs\" arguments use a subset of the following field abbreviations:"
      "  x            completed"
      "  pri or prio  priority"
      "  cod          completion date (\"co\" for completion, \"d\" for date)"
      "  crd          creation date (\"cr\" for creation, \"d\" for date)"
      "  desc         description"
      "  @            context (contexts are marked with @)"
      "  +            project (projects are marked with +)"
      "  <key>:       tag key, say, \"due:\" (tags are marked as key:value)"
      ""
      "Group and sort specs are of the form  <order><field-abbreviation> , where"
      "<order> is \"+\" or \"-\" for ascending or descending order, respectively."
      "<order> can be omitted, in which case ascending order is assumed. Several"
      "field abbreviations can be combined by separating them with commas (but"
      "NOT spaces after the commas), for example \"x,pri\"."
      ""
      "The valid field abbreviation subsets for the different specs are:"
      "  Sort specs: x, pri, prio, cod, crd, desc"
      "  Group specs: x, pri, prio, cod, crd, desc, @, +, <key>:"
      "  Field omission specs: x, pri, prio, cod, crd, auto (\"auto\" means that"
      "    field omissions are added from the group specs). The idea is that if you"
      "    group by, say, priority, listing the priority for each task is redundant."
      ""
      "Group specs example:"
      "  pri,+  Group by ascending priority. Within each such group, group by"
      "  ascending project name."
      ""
      "Sort specs example:"
      "  -due:,desc  Sort first by descending due date. Sort tasks with the same due"
      "    date by description in ascending order."
      ""
      "Field omission specs example:"
      "  x,crd  For each task in the output, omit the completion status (the \"x\""
      "    prefix) and the creation date. As a special rule, omitting the creation"
      "    date also omits the completion date. This is because only omitting the"
      "    creation date might leave a completion date, but if there's a single date"
      "    on a line, it's supposed to be the creation date!"
      ""
      "After the options, specify one or more paths of todo.txt-format files. As a"
      "special case, if no file path is given, todoreport tries to get the path from"
      "the environment variable TODO_FILE ."
      ""
      "Example:"
      "  todoreport -g +,due: -s pri,-crd,desc -o auto my_todo.txt another_todo.txt"
      ""
      "See the project website for more: https://git.sr.ht/~sschwarzer/todo-txt"
    #:args args
    (set! todo-file-paths-pre (map expand-user-path args)))
  (define todo-file-paths
    (cond
      [(list? todo-file-paths-pre) todo-file-paths-pre]
      [(getenv TODO-FILE-ENV-NAME) (list (getenv TODO-FILE-ENV-NAME))]
      [else (raise-user-error "no paths and no environment variable `TODO_FILE`")]))
  (define all-tasks
    (append* (map file->tasks todo-file-paths)))
  (define group-specs
    (let* ([specs (specs-string->specs group-specs-string 'group)]
           ; Evaluated for possible side effect.
           [_ (check-specs specs "group specs")])
      (clean-specs specs)))
  (define sort-specs
    (let* ([specs (specs-string->specs sort-specs-string 'sort)]
           ; Evaluated for possible side effect.
           [_ (check-specs specs "sort specs")])
      (clean-specs specs)))
  (define field-omission-specs
    (let* ([initial-specs (specs-string->specs field-omission-specs-string 'omit-fields)]
           [specs (update-field-omission-specs initial-specs group-specs)]
           ; Evaluated for possible side effect.
           [_ (check-specs specs "field omission specs")])
      (clean-specs specs)))
  (command-line-data all-tasks group-specs sort-specs field-omission-specs))


; --------------------------------------------------------------------------------

(module+ test
  (require
    rackunit
    al2-test-runner
    (submod ".."))

  ; Return curried version of `specs-string->specs` to avoid passing the
  ; `spec-type`.
  (define (custom-specs-string->specs spec-type)
    (lambda (specs-string)
      (specs-string->specs specs-string spec-type)))

  (define sort-specs-string->specs        (custom-specs-string->specs 'sort))
  (define group-specs-string->specs       (custom-specs-string->specs 'group))
  (define omit-fields-specs-string->specs (custom-specs-string->specs 'omit-fields))

  (run-tests
    (test-suite "spec-string->spec for sort specs"
      (test-equal? "Empty string"
        (sort-specs-string->specs "")
        '())
      (test-equal? "Completion"
        (sort-specs-string->specs "x")
        (list (task-sort-spec 'asc 'completed?)))
      (test-equal? "Priority without indicator"
        (sort-specs-string->specs "pri")
        (list (task-sort-spec 'asc 'priority)))
      (test-equal? "Priority with ascending indicator"
        (sort-specs-string->specs "+pri")
        (list (task-sort-spec 'asc 'priority)))
      (test-equal? "Priority with descending indicator"
        (sort-specs-string->specs "-pri")
        (list (task-sort-spec 'desc 'priority)))
      (test-equal? "Priority from \"prio\""
        (sort-specs-string->specs "prio")
        (list (task-sort-spec 'asc 'priority)))
      (test-equal? "Completion date"
        (sort-specs-string->specs "cod")
        (list (task-sort-spec 'asc 'completion-date)))
      (test-equal? "Creation date"
        (sort-specs-string->specs "crd")
        (list (task-sort-spec 'asc 'creation-date)))
      (test-equal? "Description"
        (sort-specs-string->specs "desc")
        (list (task-sort-spec 'asc 'description)))
      (test-equal? "Multiple specs"
        (sort-specs-string->specs "+prio,-cod")
        (list (task-sort-spec 'asc 'priority) (task-sort-spec 'desc 'completion-date)))
      (test-equal? "Invalid spec"
        (sort-specs-string->specs "invalid")
        '("invalid"))
      (test-equal? "Can't sort by contexts"
        (sort-specs-string->specs "@")
        '("@"))
      (test-equal? "Can't sort by projects"
        (sort-specs-string->specs "+")
        '("+"))
      (test-equal? "Can't sort by tag"
        (sort-specs-string->specs "due:")
        '("due:"))
      (test-equal? "Redundant prefixes"
        (sort-specs-string->specs "+-pri")
        '("+-pri"))
      (test-equal? "Multiple specs including invalid spec"
        (sort-specs-string->specs "+prio,invalid,-cod")
        (list (task-sort-spec 'asc 'priority)
              "invalid"
              (task-sort-spec 'desc 'completion-date)))
    ))

  (run-tests
    (test-suite "specs-string->specs for group specs"
      (test-equal? "Empty string"
        (group-specs-string->specs "")
        '())
      (test-equal? "Completion"
        (group-specs-string->specs "x")
        (list (task-group-spec 'asc 'completed? #f)))
      (test-equal? "Priority without indicator"
        (group-specs-string->specs "pri")
        (list (task-group-spec 'asc 'priority #f)))
      (test-equal? "Priority with ascending indicator"
        (group-specs-string->specs "+pri")
        (list (task-group-spec 'asc 'priority #f)))
      (test-equal? "Priority with descending indicator"
        (group-specs-string->specs "-pri")
        (list (task-group-spec 'desc 'priority #f)))
      (test-equal? "Priority from \"prio\""
        (group-specs-string->specs "prio")
        (list (task-group-spec 'asc 'priority #f)))
      (test-equal? "Completion date"
        (group-specs-string->specs "cod")
        (list (task-group-spec 'asc 'completion-date #f)))
      (test-equal? "Creation date"
        (group-specs-string->specs "crd")
        (list (task-group-spec 'asc 'creation-date #f)))
      (test-equal? "Description"
        (group-specs-string->specs "desc")
        (list (task-group-spec 'asc 'description #f)))
      (test-equal? "Contexts"
        (group-specs-string->specs "@")
        (list (task-group-spec 'asc 'contexts #f)))
      (test-equal? "Projects"
        (group-specs-string->specs "+")
        (list (task-group-spec 'asc 'projects #f)))
      (test-equal? "Tags"
        (group-specs-string->specs "due:")
        (list (task-group-spec 'asc 'tags "due")))
      (test-equal? "Multiple specs"
        (group-specs-string->specs "+prio,-cod,++,-+")
        (list (task-group-spec 'asc 'priority #f)
              (task-group-spec 'desc 'completion-date #f)
              (task-group-spec 'asc 'projects #f)
              (task-group-spec 'desc 'projects #f)))
      (test-equal? "Invalid spec"
        (group-specs-string->specs "invalid")
        '("invalid"))
      (test-equal? "Redundant prefixes"
        (group-specs-string->specs "+-pri")
        '("+-pri"))
      (test-equal? "Multiple specs including invalid spec"
        (group-specs-string->specs "+prio,invalid,-cod")
        (list (task-group-spec 'asc 'priority #f)
              "invalid"
              (task-group-spec 'desc 'completion-date #f)))
    ))

  (run-tests
    (test-suite "spec-string->spec for fields to omit"
      (test-equal? "Empty string"
        (omit-fields-specs-string->specs "")
        '())
      (test-equal? "Completion"
        (omit-fields-specs-string->specs "x")
        '(completed?))
      (test-equal? "Priority"
        (omit-fields-specs-string->specs "pri")
        '(priority))
      (test-equal? "Priority from \"prio\""
        (omit-fields-specs-string->specs "prio")
        '(priority))
      (test-equal? "Completion date"
        (omit-fields-specs-string->specs "cod")
        '(completion-date))
      (test-equal? "Creation date"
        (omit-fields-specs-string->specs "crd")
        '(creation-date))
      (test-equal? "Multiple specs"
        (omit-fields-specs-string->specs "prio,cod")
        '(priority completion-date))
      (test-equal? "Invalid spec"
        (omit-fields-specs-string->specs "invalid")
        '("invalid"))
      (test-case "No order indicator allowed"
        (check-equal?
          (omit-fields-specs-string->specs "+pri")
          '("+pri"))
        (check-equal?
          (omit-fields-specs-string->specs "-pri")
          '("-pri")))
      (test-equal? "Can't omit description"
        (omit-fields-specs-string->specs "desc")
        '("desc"))
      (test-case "Can't omit complex fields"
        (check-equal?
          (omit-fields-specs-string->specs "@")
          '("@"))
        (check-equal?
          (omit-fields-specs-string->specs "+")
          '("+"))
        (check-equal?
          (omit-fields-specs-string->specs "due:")
          '("due:")))
      (test-equal? "Multiple specs including invalid spec"
        (omit-fields-specs-string->specs "prio,invalid,cod")
        '(priority "invalid" completion-date))
    ))

  (run-tests
    (test-suite "Spec fields processing"
      ; === `update-field-omission-specs`
      ; Field omission specs should stay as before.
      (test-equal? "No update necessary because of no dates and no \"auto\" field"
        (update-field-omission-specs '(priority completed?)
                                     (list (task-group-spec 'asc 'priority #f)
                                           (task-group-spec 'asc 'contexts #f)))
        '(priority completed?))
      ; "auto" should include fields from `task-group-spec`s, but only those
      ; fields that make sense for field omission specs.
      (test-equal? "Update of \"auto\" field"
        (update-field-omission-specs '(priority "auto")
                                     (list (task-group-spec 'asc 'completed? #f)
                                           (task-group-spec 'asc 'contexts #f)))
        '(priority completed?))
      ; If the creation date should be omitted, the completion date must be
      ; omitted, too, because a single date _must_ be the creation date.
      (test-equal? "Add completion date"
        (update-field-omission-specs '(priority creation-date)
                                     (list (task-group-spec 'asc 'completed? #f)
                                           (task-group-spec 'asc 'contexts #f)))
        '(priority completion-date creation-date))
      ; If only the completion date should be omitted, that's fine because a
      ; single remaining date can be a creation date.
      (test-equal? "No date to add"
        (update-field-omission-specs '(priority completion-date)
                                     (list (task-group-spec 'asc 'completed? #f)
                                           (task-group-spec 'asc 'contexts #f)))
        '(priority completion-date))
      ; If the creation date comes from the `task-group-spec`s, the completion
      ; date should also be added.
      (test-equal? "Add completion date for creation date from group specs"
        (update-field-omission-specs '(priority "auto")
                                     (list (task-group-spec 'asc 'creation-date #f)
                                           (task-group-spec 'asc 'contexts #f)))
        '(priority completion-date creation-date))
      ; Addition of a field that's already there is ok. The duplicated fields
      ; will be removed in a later step when processing the omission specs.
      (test-equal? "Redundant fields"
        (update-field-omission-specs '(priority "auto" creation-date)
                                     (list (task-group-spec 'asc 'priority #f)
                                           (task-group-spec 'asc 'completion-date #f)))
        ; No need to add another completion date because it was already added
        ; via the group specs.
        '(priority priority completion-date creation-date))
      ; === `check-specs`
      (test-case "`check-specs` should succeed without strings in list"
        (check-specs '(priority description) "sort specs"))
      (test-exn "String in specs should raise user error"
        (regexp "^todoreport: invalid sort specs: foo bar$")
        (lambda ()
          (check-specs '(priority "foo" 'description "bar") "sort specs")))
      ; === `clean-specs`
      (test-equal? "Duplicates should be removed"
        (clean-specs
          '(priority description priority completion-date description creation-date))
        '(priority description completion-date creation-date))
    ))
)
